import os
import pytest
import unittest
from typing import List
from kservehelper.model import KServeModel
from kservehelper.types import Input, Path


class CustomModel:

    def load(self):
        pass

    def predict(
            self,
            angle: int = Input(
                description="Rotation angle",
                default=90
            )
    ) -> List[Path]:
        from PIL import Image

        paths = []
        folder = os.path.dirname(os.path.abspath(__file__))
        input_image = Image.open(os.path.join(folder, "dog.jpg"))
        for i in range(4):
            image = input_image.rotate(angle + i * 10)
            output_path = f"/tmp/dog_{i}.jpg"
            image.save(output_path)
            paths.append(Path(output_path))
        return paths


class TestKServeModel(unittest.TestCase):

    @pytest.mark.skip
    def test_model(self):
        payload = {"angle": 45, "upload_webhook": "http://localhost:12000/upload"}
        model = KServeModel("test", CustomModel)
        outputs = model.predict(payload)
        print(outputs)


if __name__ == "__main__":
    unittest.main()



import json
import base64
import io
import requests
from PIL import Image


def test():
    url = "http://localhost:8080/v1/models/stable-diffusion:predict"
    with open("input.json", "r") as f:
        payload = json.load(f)
    response = requests.post(url, json=payload)
    r = response.json()

    im_binary = base64.b64decode(r["image"])
    image = Image.open(io.BytesIO(im_binary))
    image.save("test.png", format="PNG")


if __name__ == "__main__":
    test()



import base64
import io
import torch
from typing import Dict
from torch import autocast
from diffusers import StableDiffusionPipeline
from kservehelper import KServeModel
from kservehelper.types import Input


class StableDiffusion:

    def __init__(self):
        self.device = None
        self.text2img_pipe = None

    def load(self):
        model_path = "/mnt/models/stable-diffusion/v1_4"
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.text2img_pipe = StableDiffusionPipeline.from_pretrained(
            model_path).to(self.device)

    def predict(
            self,
            prompt: str = Input(
                description="Input prompt",
                default="a dog playing on the street"
            ),
            guidance_scale: float = Input(
                description="Guidance scale",
                default=7.5
            ),
            height: int = Input(
                description="Image height",
                default=512
            ),
            width: int = Input(
                description="Image width",
                default=512
            ),
            num_inference_steps: int = Input(
                description="The number of inference steps",
                default=50
            ),
            seed: int = Input(
                description="Random seed",
                default=12345
            ),
            safety_check: bool = Input(
                description="Do safety check",
                default=False
            )
    ) -> Dict:
        if not safety_check:
            self.text2img_pipe.safety_checker = lambda images, **kwargs: (images, False)
        generator = torch.Generator(self.device)
        generator.manual_seed(seed)

        with autocast(self.device):
            image = self.text2img_pipe(
                prompt=prompt,
                guidance_scale=guidance_scale,
                height=height,
                width=width,
                num_inference_steps=num_inference_steps,
                generator=generator
            ).images[0]

        buffer = io.BytesIO()
        image.save(buffer, format="PNG")
        im_b64 = base64.b64encode(buffer.getvalue())
        return {"image": im_b64}


if __name__ == "__main__":
    KServeModel.serve("stable-diffusion", StableDiffusion)



import os
import yaml
import logging
import subprocess
from queue import Queue
from threading import Thread
from typing import Dict, Sequence
from schema import Schema, Optional, SchemaError

logging.basicConfig(level="INFO")
logger = logging.getLogger("Docker build")

SCHEMA = Schema(
    {
        "build": {
            Optional("cuda"): lambda x: x in ["11.7", "12.1"],
            Optional("python_version"): lambda x: x in ["3.8", "3.9", "3.10"],
            Optional("cuda_image_type"): lambda x: x in ["base", "runtime", "devel"],
            Optional("cudnn"): bool,
            Optional("system_packages"): list,
            Optional("python_packages"): list,
        },
        "image": str,
        "entrypoint": str
    }
)

DOCKER_IMAGES = {
    "default": "ubuntu22.04",
    "python:3.8": "python:3.8.18-slim",
    "python:3.9": "python:3.9.18-slim",
    "python:3.10": "python:3.10.13-slim",
    "cuda:11.7,python:3.8": "nvidia/cuda:11.7.1-base-ubuntu20.04",
    "cuda:12.1,python:3.8": "nvidia/cuda:12.1.0-base-ubuntu20.04",
    "cuda:11.7,python:3.10": "nvidia/cuda:11.7.1-base-ubuntu22.04",
    "cuda:12.1,python:3.10": "nvidia/cuda:12.1.0-base-ubuntu22.04",
}
CONFIGFILE = "config.yaml"
DOCKERFILE = "Dockerfile"

IGNORED_FILES = [
    "__pycache__",
    CONFIGFILE,
    DOCKERFILE
]


def _get_cuda_image(cuda_version: str, python_version: str, image_type: str, cudnn: bool):
    assert image_type in ["base", "runtime", "devel"]
    cudas = {
        "11.7": "11.7.1",
        "12.1": "12.1.0"
    }
    pythons = {
        "3.8": "ubuntu20.04",
        "3.10": "ubuntu22.04"
    }
    parts = [f"nvidia/cuda:{cudas.get(cuda_version, cuda_version)}"]
    if cudnn:
        parts.append("cudnn8")
    parts.append(image_type)
    parts.append(pythons.get(python_version, "ubuntu22.04"))
    return "-".join(parts)


def _load_config(filepath: str) -> Dict:
    with open(filepath, "r") as f:
        config = yaml.safe_load(f)
    try:
        SCHEMA.validate(config)
        return config
    except SchemaError as e:
        raise RuntimeError(
            "The image config does not fit the required schema."
        ) from e


def _write_dockerfile(config: Dict, folder: str):
    # Choose the base image
    python_installed = False
    if "cuda" in config["build"]:
        base_image = _get_cuda_image(
            cuda_version=config["build"]["cuda"],
            python_version=config["build"].get("python_version", "3.10"),
            image_type=config["build"].get("cuda_image_type", "base"),
            cudnn=config["build"].get("cudnn", False),
        )
    elif "python_version" in config["build"]:
        tag = f"python:{config['build']['python_version']}"
        base_image = DOCKER_IMAGES[tag]
        python_installed = True
    else:
        base_image = DOCKER_IMAGES["default"]

    commands = [
        f"FROM {base_image}",
        "ENV DEBIAN_FRONTEND=noninteractive",
        "RUN apt update && apt upgrade -y",
        ""
    ]
    # Install system packages
    if "system_packages" in config["build"] and len(config["build"]["system_packages"]) > 0:
        commands.append("RUN apt install -y --no-install-recommends \\")
        for package in config["build"]["system_packages"][:-1]:
            commands.append(f"    {package} \\")
        commands.append(f"    {config['build']['system_packages'][-1]}")
        commands.append("")

    # Install default python
    if not python_installed:
        commands.append("RUN apt install -y python3 python3-pip")
        commands.append("")

    # Install python dependencies
    if "python_packages" in config["build"] and len(config["build"]["python_packages"]) > 0:
        commands.append("RUN pip install --no-cache-dir --upgrade pip")
        commands.append("RUN pip install --no-cache-dir \\")
        for package in config["build"]["python_packages"][:-1]:
            commands.append(f"    {package} \\")
        commands.append(f"    {config['build']['python_packages'][-1]}")
        commands.append("")

    # Copy the code and artifacts
    commands.append("WORKDIR /app")
    commands.append("")
    for file in os.listdir(folder):
        if file not in IGNORED_FILES:
            commands.append(f"COPY {file} .")
    commands.append("")
    commands.append(f'CMD ["python3", "/app/{config["entrypoint"]}"]')

    with open(os.path.join(folder, DOCKERFILE), "w") as f:
        for command in commands:
            f.write(f"{command}\n")


# Adapted from BentoML: https://github.com/bentoml/BentoML/blob/main/src/bentoml/_internal/container/base.py
def _stream_logs(cmds: Sequence[str], *, env=None, cwd=None):
    proc = subprocess.Popen(
        cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env, cwd=cwd
    )
    queue: Queue[tuple[str, bytes]] = Queue()
    stderr, stdout = b"", b""
    # We will use a thread to read from the subprocess and avoid hanging from Ctrl+C
    t = Thread(target=_enqueue_output, args=(proc.stdout, "stdout", queue))
    t.daemon = True
    t.start()
    t = Thread(target=_enqueue_output, args=(proc.stderr, "stderr", queue))
    t.daemon = True
    t.start()
    for _ in range(2):
        for src, line in iter(queue.get, None):
            logger.info(line.decode(errors="replace").strip("\n"))
            if src == "stderr":
                stderr += line
            else:
                stdout += line
    exit_code = proc.wait()
    if exit_code != 0:
        raise subprocess.CalledProcessError(
            exit_code, cmds, output=stdout, stderr=stderr
        )
    return subprocess.CompletedProcess(
        proc.args, exit_code, stdout=stdout, stderr=stderr
    )


def _enqueue_output(pipe, pipe_name, queue):
    try:
        with pipe:
            for line in iter(pipe.readline, b""):
                queue.put((pipe_name, line))
    finally:
        queue.put(None)


def build(folder: str, quiet: bool = False):
    # Build Dockerfile
    config = _load_config(os.path.join(folder, CONFIGFILE))
    _write_dockerfile(config, folder)
    # Build docker image
    try:
        command = ["docker", "build", "--platform=linux/amd64"]
        if quiet:
            command.append("--quiet")
        command += ["-t", config["image"], folder]
        _stream_logs(command)
    except subprocess.CalledProcessError as e:
        print(e)


def push(folder: str):
    config = _load_config(os.path.join(folder, CONFIGFILE))
    try:
        command = ["docker", "push", config["image"]]
        _stream_logs(command)
    except subprocess.CalledProcessError as e:
        print(e)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import socket
from typing import Dict, List, Optional, Union

import pkg_resources
import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.responses import ORJSONResponse
from fastapi.routing import APIRoute as FastAPIRoute
from prometheus_client import REGISTRY, exposition
from timing_asgi import TimingClient, TimingMiddleware
from timing_asgi.integrations import StarletteScopeToName

from kserve.errors import (InferenceError, InvalidInput, ModelNotFound,
                           ModelNotReady, generic_exception_handler,
                           inference_error_handler, invalid_input_handler,
                           model_not_found_handler, model_not_ready_handler,
                           not_implemented_error_handler)
from kserve.logging import trace_logger
from kserve.protocol.dataplane import DataPlane

from kservehelper.kserve.rest.v1_endpoints import V1Endpoints
from kserve.protocol.rest.v2_datamodels import (InferenceResponse, ModelMetadataResponse,
                                                ModelReadyResponse, ServerLiveResponse,
                                                ServerMetadataResponse, ServerReadyResponse)
from kserve.protocol.rest.v2_endpoints import V2Endpoints


async def metrics_handler(request: Request) -> Response:
    encoder, content_type = exposition.choose_encoder(request.headers.get("accept"))
    return Response(content=encoder(REGISTRY), headers={"content-type": content_type})


class PrintTimings(TimingClient):
    def timing(self, metric_name, timing, tags):
        trace_logger.info(f"{metric_name}: {timing}")


class _NoSignalUvicornServer(uvicorn.Server):
    def install_signal_handlers(self) -> None:
        pass


class RESTServer:
    def __init__(self, data_plane: DataPlane, model_repository_extension, enable_docs_url=False):
        self.dataplane = data_plane
        self.model_repository_extension = model_repository_extension
        self.enable_docs_url = enable_docs_url

    def create_application(self) -> FastAPI:
        """Create a KServe ModelServer application with API routes.

        Returns:
            FastAPI: An application instance.
        """
        v1_endpoints = V1Endpoints(self.dataplane, self.model_repository_extension)
        v2_endpoints = V2Endpoints(self.dataplane, self.model_repository_extension)

        return FastAPI(
            title="KServe ModelServer",
            version=pkg_resources.get_distribution("kserve").version,
            docs_url="/docs" if self.enable_docs_url else None,
            redoc_url=None,
            default_response_class=ORJSONResponse,
            routes=[
                # Server Liveness API returns 200 if server is alive.
                FastAPIRoute(r"/", self.dataplane.live),
                # Metrics
                FastAPIRoute(r"/metrics", metrics_handler, methods=["GET"]),
                # V1 Inference Protocol
                FastAPIRoute(r"/v1/models", v1_endpoints.models, tags=["V1"]),
                # Model Health API returns 200 if model is ready to serve.
                FastAPIRoute(r"/v1/models/{model_name}", v1_endpoints.model_ready, tags=["V1"]),
                FastAPIRoute(r"/v1/models/{model_name}:predict",
                             v1_endpoints.predict, methods=["POST"], tags=["V1"]),
                FastAPIRoute(r"/v1/models/{model_name}:explain",
                             v1_endpoints.explain, methods=["POST"], tags=["V1"]),
                # Model docs
                FastAPIRoute(r"/v1/docs/{model_name}",
                             v1_endpoints.docs, methods=["GET"], tags=["V1"]),
                # V2 Inference Protocol
                # https://github.com/kserve/kserve/tree/master/docs/predict-api/v2
                FastAPIRoute(r"/v2", v2_endpoints.metadata,
                             response_model=ServerMetadataResponse, tags=["V2"]),
                FastAPIRoute(r"/v2/health/live", v2_endpoints.live,
                             response_model=ServerLiveResponse, tags=["V2"]),
                FastAPIRoute(r"/v2/health/ready", v2_endpoints.ready,
                             response_model=ServerReadyResponse, tags=["V2"]),
                FastAPIRoute(r"/v2/models/{model_name}",
                             v2_endpoints.model_metadata, response_model=ModelMetadataResponse, tags=["V2"]),
                FastAPIRoute(r"/v2/models/{model_name}/versions/{model_version}",
                             v2_endpoints.model_metadata, tags=["V2"], include_in_schema=False),
                FastAPIRoute(r"/v2/models/{model_name}/ready",
                             v2_endpoints.model_ready, response_model=ModelReadyResponse, tags=["V2"]),
                FastAPIRoute(r"v2/models/{model_name}/versions/{model_version}/ready",
                             v2_endpoints.model_ready, response_model=ModelReadyResponse, tags=["V2"]),
                FastAPIRoute(r"/v2/models/{model_name}/infer",
                             v2_endpoints.infer, methods=["POST"], response_model=InferenceResponse, tags=["V2"]),
                FastAPIRoute(r"/v2/models/{model_name}/versions/{model_version}/infer",
                             v2_endpoints.infer, methods=["POST"], tags=["V2"], include_in_schema=False),
                FastAPIRoute(r"/v2/repository/models/{model_name}/load",
                             v2_endpoints.load, methods=["POST"], tags=["V2"]),
                FastAPIRoute(r"/v2/repository/models/{model_name}/unload",
                             v2_endpoints.unload, methods=["POST"], tags=["V2"]),
            ], exception_handlers={
                InvalidInput: invalid_input_handler,
                InferenceError: inference_error_handler,
                ModelNotFound: model_not_found_handler,
                ModelNotReady: model_not_ready_handler,
                NotImplementedError: not_implemented_error_handler,
                Exception: generic_exception_handler
            }
        )


class UvicornServer:
    def __init__(self, http_port: int, sockets: List[socket.socket],
                 data_plane: DataPlane, model_repository_extension, enable_docs_url,
                 log_config: Optional[Union[str, Dict]] = None,
                 access_log_format: Optional[str] = None):
        super().__init__()
        self.sockets = sockets
        rest_server = RESTServer(data_plane, model_repository_extension, enable_docs_url)
        app = rest_server.create_application()
        app.add_middleware(
            TimingMiddleware,
            client=PrintTimings(),
            metric_namer=StarletteScopeToName(prefix="kserve.io", starlette_app=app)
        )
        self.cfg = uvicorn.Config(
            app=app,
            host="0.0.0.0",
            port=http_port,
            log_config=log_config
        )

        # More context in https://github.com/encode/uvicorn/pull/947
        # At the time of writing the ASGI specs are not clear when it comes
        # to change the access log format, and hence the Uvicorn upstream devs
        # chose to create a custom middleware for this.
        # The allowed log format is specified in https://github.com/Kludex/asgi-logger#usage
        if access_log_format:
            from asgi_logger import AccessLoggerMiddleware

            # As indicated by the asgi-logger docs, we need to clear/unset
            # any setting for uvicorn.access to avoid log duplicates.
            logging.getLogger("uvicorn.access").handlers = []
            app.add_middleware(
                AccessLoggerMiddleware, format=access_log_format)
            # The asgi-logger settings don't set propagate to False,
            # so we get duplicates if we don't set it explicitly.
            logging.getLogger("access").propagate = False

        self.server = _NoSignalUvicornServer(config=self.cfg)

    def run_sync(self):
        server = uvicorn.Server(config=self.cfg)
        asyncio.run(server.serve(sockets=self.sockets))

    async def run(self):
        await self.server.serve()

    async def stop(self, sig: Optional[int] = None):
        if self.server:
            self.server.handle_exit(sig=sig, frame=None)



import unittest
from typing import Dict
from kservehelper.model import KServeModel
from kservehelper.types import Input


class CustomModel:

    def load(self):
        pass

    def predict(
            self,
            prompt: str = Input(
                description="Input prompt",
                default="standing, (full body)++"
            )
    ) -> Dict:
        return {"outputs": prompt}


class TestKServeModel(unittest.TestCase):

    def test_model(self):
        payload = {"prompt": "test test", "upload_webhook": "http://localhost"}
        model = KServeModel("test", CustomModel)
        outputs = model.predict(payload)
        self.assertEqual(outputs["outputs"], "test test")

    def test_default(self):
        payload = {"upload_webhook": "http://localhost"}
        model = KServeModel("test", CustomModel)
        outputs = model.predict(payload)
        self.assertEqual(outputs["outputs"], "standing, (full body)++")


if __name__ == "__main__":
    unittest.main()



import os
from PIL import Image
from kservehelper import KServeModel
from kservehelper.types import Input, Path


class Model:

    def load(self):
        pass

    def predict(
            self,
            angle: int = Input(
                description="Rotation angle",
                default=90
            )
    ) -> Path:
        folder = os.path.dirname(os.path.abspath(__file__))
        input_image = Image.open(os.path.join(folder, "dog.jpg"))
        image = input_image.rotate(angle)
        output_path = KServeModel.generate_filepath("dog.jpg")
        image.save(output_path)
        return Path(output_path)


if __name__ == "__main__":
    KServeModel.serve("test-rotate-image", Model)



import time
import json
import uuid
import typing
import pydantic
import inspect

from collections import OrderedDict
from typing import Any, Dict, Callable
from inspect import signature

from kserve import Model
from kservehelper.kserve import ModelServer
from kservehelper.types import Path
from kservehelper.utils import upload_files


class ModelIOInfo:

    def __init__(self):
        self._input_info = None
        self._output_info = None

    @staticmethod
    def _fieldinfo2dict(info):
        d = OrderedDict()
        for arg in info.__repr_args__():
            if arg[0] != "extra":
                d[arg[0]] = arg[1]
            else:
                d.update({k: v for k, v in arg[1].items() if v})
        return d

    def set_input_signatures(self, method: Callable):
        t = signature(method)
        input_info = OrderedDict()
        for key, value in t.parameters.items():
            if isinstance(value.default, pydantic.fields.FieldInfo):
                d = ModelIOInfo._fieldinfo2dict(value.default)
                if value.annotation != inspect._empty:
                    d["type"] = value.annotation.__name__
                input_info[key] = d
        self._input_info = input_info

    def set_output_signatures(self, method: Callable):
        t = signature(method)
        output_info = OrderedDict()
        if t.return_annotation == inspect._empty:
            output_info["type"] = None
            output_info["args"] = ()
        else:
            output_info["type"] = typing.get_origin(t.return_annotation)
            if output_info["type"] is None:
                output_info["type"] = t.return_annotation
            output_info["args"] = typing.get_args(t.return_annotation)
        self._output_info = output_info

    @property
    def inputs(self):
        return self._input_info

    @property
    def outputs(self):
        return self._output_info


class KServeModel(Model):
    MODEL_IO_INFO = ModelIOInfo()
    HAS_PREPROCESS = False
    HAS_PREDICT = False
    HAS_POSTPROCESS = False

    def __init__(self, name: str, model_class: Any):
        super().__init__(name)
        KServeModel._reset()
        KServeModel._build_functions(model_class)
        self.model = model_class()

        # Only used for transforms
        self.upload_webhook = None
        self.predict_start_time = -1

    def load(self) -> bool:
        method = getattr(self.model, "load", None)
        if callable(method):
            self.model.load()
        self.ready = True
        return self.ready

    def docs(self):
        return json.loads(json.dumps(KServeModel.MODEL_IO_INFO.inputs))

    @staticmethod
    def _build_functions(model_class):
        # Predict function
        method = getattr(model_class, "predict", None)
        if callable(method):
            KServeModel.MODEL_IO_INFO.set_input_signatures(method)
            KServeModel.MODEL_IO_INFO.set_output_signatures(method)
            setattr(KServeModel, "predict", KServeModel._predict)
            KServeModel.HAS_PREDICT = True

        # Preprocess function
        method = getattr(model_class, "preprocess", None)
        if callable(method):
            assert not KServeModel.HAS_PREDICT, \
                "`predict` function has been defined already, cannot define `preprocess` separately"
            KServeModel.MODEL_IO_INFO.set_input_signatures(method)
            setattr(KServeModel, "preprocess", KServeModel._preprocess)
            KServeModel.HAS_PREPROCESS = True

        # Postprocess function
        method = getattr(model_class, "postprocess", None)
        if callable(method):
            assert not KServeModel.HAS_PREDICT, \
                "`predict` function has been defined already, cannot define `postprocess` separately"
            KServeModel.MODEL_IO_INFO.set_output_signatures(method)
            setattr(KServeModel, "postprocess", KServeModel._postprocess)
            KServeModel.HAS_POSTPROCESS = True

        if KServeModel.HAS_PREPROCESS or KServeModel.HAS_POSTPROCESS:
            assert KServeModel.HAS_PREPROCESS and KServeModel.HAS_POSTPROCESS, \
                "`preprocess` and `postprocess` must be both defined"

    @staticmethod
    def _process_payload(payload: Dict):
        for key, values in KServeModel.MODEL_IO_INFO.inputs.items():
            if key not in payload:
                payload[key] = values["default"]
        return payload

    @staticmethod
    def _predict(self, payload: Dict, headers: Dict[str, str] = None) -> Dict:
        start_time = time.time()
        upload_webhook = payload.pop("upload_webhook", None)
        payload = KServeModel._process_payload(payload)
        outputs = self.model.predict(**payload)
        results = KServeModel._upload(upload_webhook, outputs)
        results["running_time"] = f"{time.time() - start_time}s"
        return results

    @staticmethod
    def _preprocess(self, payload: Dict, headers: Dict[str, str] = None) -> Dict:
        self.predict_start_time = time.time()
        self.upload_webhook = payload.pop("upload_webhook", None)
        payload = KServeModel._process_payload(payload)
        return self.model.preprocess(**payload)

    @staticmethod
    def _postprocess(self, infer_response: Dict, headers: Dict[str, str] = None) -> Dict:
        outputs = self.model.postprocess(infer_response)
        results = KServeModel._upload(self.upload_webhook, outputs)
        results["running_time"] = f"{time.time() - self.predict_start_time}s"
        self.upload_webhook = None
        return results

    @staticmethod
    def _upload(upload_webhook, model_outputs):
        if KServeModel.MODEL_IO_INFO.outputs is None:
            assert isinstance(model_outputs, dict), "Model output must be a dict"
            return model_outputs

        if KServeModel.MODEL_IO_INFO.outputs["type"] == Path:
            assert upload_webhook is not None, \
                "Model output type is `Path`, but `upload_webhook` is not set"
            assert isinstance(model_outputs, Path), \
                "Model output type is `Path`, but the actual output is not `Path`"
            return upload_files(upload_webhook, [model_outputs])

        if KServeModel.MODEL_IO_INFO.outputs["type"] == list:
            if len(KServeModel.MODEL_IO_INFO.outputs["args"]) == 1 and \
                    KServeModel.MODEL_IO_INFO.outputs["args"][0] == Path:
                assert upload_webhook is not None, \
                    "Model output type is `Path`, but `upload_webhook` is not set"
                assert isinstance(model_outputs, (list, tuple)), \
                    "Model output type is `List[Path]`, but the actual output is not a List"
                return upload_files(upload_webhook, model_outputs)

        assert isinstance(model_outputs, dict), "Model output must be a dict"
        return model_outputs

    @staticmethod
    def _reset():
        KServeModel.MODEL_IO_INFO = ModelIOInfo()
        if KServeModel.HAS_PREPROCESS:
            KServeModel.HAS_PREPROCESS = False
            method = getattr(KServeModel, "preprocess", None)
            if callable(method):
                delattr(KServeModel, "preprocess")
        if KServeModel.HAS_PREDICT:
            KServeModel.HAS_PREDICT = False
            method = getattr(KServeModel, "predict", None)
            if callable(method):
                delattr(KServeModel, "predict")
        if KServeModel.HAS_POSTPROCESS:
            KServeModel.HAS_POSTPROCESS = False
            method = getattr(KServeModel, "postprocess", None)
            if callable(method):
                delattr(KServeModel, "postprocess")

    @staticmethod
    def generate_filepath(filename: str) -> str:
        return f"/tmp/{str(uuid.uuid4())}-{filename}"

    @staticmethod
    def serve(name: str, model_class: Any):
        model = KServeModel(name, model_class)
        ModelServer().start([model])



from setuptools import setup, find_namespace_packages

setup(
    name="kservehelper",
    version="1.0.7",
    author="Wenzhuo Yang",
    description="A KServe Model Wrapper",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/yangwenz/kserve-helper",
    license="3-Clause BSD",
    packages=find_namespace_packages(include="kservehelper.*"),
    package_dir={"kservehelper": "kservehelper"},
    entry_points={
        "console_scripts": [
            "kservehelper=kservehelper.cli:main",
        ]
    },
    install_requires=[
        "kserve==0.10.2",
        "pydantic==1.10.12",
        "requests==2.29.0",
        "aiohttp==3.8.3",
        "aiofiles==23.2.1",
        "schema",
        "pyyaml",
        "click",
        "pytest"
    ],
    python_requires=">=3.8,<4",
    zip_safe=False,
)
